#pragma once

#include <algorithm>
#include <cstdint>
#include <deque>
#include <utility>
#include <vector>

namespace rmrf::net {

    class iorecord {
    private:
        size_t offset;
        std::vector<uint8_t> data;

    public:
        iorecord();
        iorecord(const void *buf, size_t size);
        iorecord(const iorecord &other);
        iorecord(iorecord &&other);

    public:
        size_t size() const;
        bool empty() const;
        void *ptr() const;

        void advance(size_t amount);
    };

    template<class iorecord_type = iorecord>
    class ioqueue {
    private:
        std::deque<iorecord_type> queue;

    public:
        static_assert(std::is_base_of<iorecord, iorecord_type>::value, "Your iorecord implementation needs to be a subclass of iorecord.");
        ioqueue() : queue{} {
            // NOP
        }

        ~ioqueue() {
            // NOP
        }

        bool empty() const {
            return this->queue.empty();
        }

        void push_back(const iorecord_type& data) {
            if (!data.empty()) {
                this->queue.push_back(data);
            }
        }
        void push_back(iorecord_type &&data) {
            if (!data.empty()) {
                this->queue.emplace_back(std::forward<iorecord_type>(data));
            }
        }

        void push_front(const iorecord_type &data) {
            if (!data.empty()) {
                this->queue.push_front(data);
            }
        }

        void push_front(iorecord_type &&data) {
            if (!data.empty()) {
                this->queue.emplace_front(std::forward<iorecord_type>(data));
            }
        }

        iorecord_type pop_front() {
            if (this->empty()) {
                return iorecord_type{};
            }

            auto result = this->queue.front();
            this->queue.pop_front();
            return result;
        }
    };

}
